part of 'package:web3dart_go_core/web3dart.dart';

class _SigningInput {
  final Transaction transaction;
  final Credentials credentials;
  final int chainId;

  _SigningInput({this.transaction, this.credentials, this.chainId});
}

Future<_SigningInput> _fillMissingData({
  @required Credentials credentials,
  @required Transaction transaction,
  int chainId,
  bool loadChainIdFromNetwork = false,
  Web3Client client,
}) async {
  assert(credentials != null);
  assert(transaction != null);
  assert(loadChainIdFromNetwork != null);
  assert(!loadChainIdFromNetwork || chainId != null,
      "You can't specify loadChainIdFromNetwork and specify a custom chain id!");

  final sender = transaction.from ?? await credentials.extractAddress();
  var energyPrice = transaction.energyPrice;
  var nonce = transaction.nonce;
  if (energyPrice == null || nonce == null) {
    if (client == null) {
      throw ArgumentError("Can't find suitable gas price and nonce from client "
          'because no client is set. Please specify a gas price on the '
          'transaction.');
    }
    energyPrice ??= await client.getGasPrice();
    nonce ??= await client.getTransactionCount(sender,
        atBlock: const BlockNum.pending());
  }

  final maxGas = transaction.maxEnergy ??
      await client
          .estimateGas(
            sender: sender,
            to: transaction.to,
            data: transaction.data,
            value: transaction.value,
            energyPrice: energyPrice,
          )
          .then((bigInt) => bigInt);

  // apply default values to null fields
  final modifiedTransaction = transaction.copyWith(
    value: transaction.value ?? BigInt.from(0),
    maxGas: maxGas,
    from: sender,
    data: transaction.data ?? Uint8List(0),
    energyPrice: energyPrice,
    nonce: nonce,
  );

  int resolvedChainId;
  if (!loadChainIdFromNetwork) {
    resolvedChainId = chainId;
  } else {
    if (client == null) {
      throw ArgumentError(
          "Can't load chain id from network when no client is set");
    }

    resolvedChainId = await client.getNetworkId();
  }

  return _SigningInput(
    transaction: modifiedTransaction,
    credentials: credentials,
    chainId: resolvedChainId,
  );
}

Future<Uint8List> _signTransaction(
    Transaction transaction, Credentials c, BigInt networkId) async {
  final _enRlp = _encodeRawToRlp(transaction, networkId);
  final _enLp = rlp.encode(_enRlp);
  final encoded = uint8ListFromList(_enLp);

  final signature = await c.signToSignature(encoded);

  final _sigEnRlp = _encodeToRlp(transaction, signature, networkId);
  final _sigLp = rlp.encode(_sigEnRlp);
  final _res = uint8ListFromList(_sigLp);

  return _res;
}

List<dynamic> _encodeToRlp(
    Transaction transaction, Uint8List signature, BigInt networkId) {
  final list = [
    transaction.nonce,
    transaction.energyPrice,
    transaction.maxEnergy,
    networkId
  ];

  if (transaction.to != null) {
    list.add(transaction.to.addressBytes);
  } else {
    list.add('');
  }

  list..add(transaction.value)..add(transaction.data);

  if (signature != null) {
    list..add(signature);
  }
  return list;
}

List<dynamic> _encodeRawToRlp(Transaction transaction, BigInt networkId) {
  // TODO: replace null values with zero
  final list = [
    transaction.nonce,
    transaction.energyPrice,
    transaction.maxEnergy,
  ];

  if (transaction.to != null) {
    list.add(transaction.to.addressBytes);
  } else {
    list.add('');
  }
  list..add(transaction.value)..add(transaction.data)..add(0);
  return list;
}
