part of 'package:web3dart_go_core/credentials.dart';

/// The sign method from ed448, so that it can be used inside [Credentials].

/// Anything that can sign payloads with a private key.
abstract class Credentials {
  static const _messagePrefix = '\u0019CoreBlockchain Signed Message:\n';

  /// Whether these [Credentials] are safe to be copied to another isolate and
  /// can operate there.
  /// If this getter returns true, the client might chose to perform the
  /// expensive signing operations on another isolate.
  bool get isolateSafe => false;

  /// Loads the ethereum address specified by these credentials.
  Future<XCBAddress> extractAddress();

  /// Signs the [payload] with a private key. The output will be like the
  /// bytes representation of the [xcb_sign RPC method](https://github.com/ethereum/wiki/wiki/JSON-RPC#xcb_sign),
  /// but without the "Ethereum signed message" prefix.
  /// The [payload] parameter contains the raw data, not a hash.
  Future<Uint8List> sign(Uint8List payload, {int chainId}) async {
    final signature = await signToSignature(payload);

    return signature;
  }

  /// Signs the [payload] with a private key and returns the obtained
  /// signature.
  Future<Uint8List> signToSignature(Uint8List payload);

  /// Signs an Core Block Chain specific signature. This method is equivalent to
  /// [sign], but with a special prefix so that this method can't be used to
  /// sign, for instance, transactions.
  Future<Uint8List> signPersonalMessage(Uint8List payload, {int chainId}) {
    final prefix = _messagePrefix + payload.length.toString();
    final prefixBytes = ascii.encode(prefix);

    // will be a Uint8List, see the documentation of Uint8List.+
    final concat = uint8ListFromList(prefixBytes + payload);

    return sign(concat, chainId: chainId);
  }
}

/// Credentials that can sign payloads with an Core Block Chain private key.
class XCBPrivateKey extends Credentials {
  final Uint8List privateKey;
  XCBAddress _cachedAddress;

  XCBPrivateKey(this.privateKey);

  XCBPrivateKey.fromHex(String hex) : privateKey = hexToBytes(hex);

  /// Creates a new, random private key from the [random] number generator.
  ///
  /// For security reasons, it is very important that the random generator used
  /// is cryptographically secure. The private key could be reconstructed by
  /// someone else otherwise. Just using [Random()] is a very bad idea! At least
  /// use [Random.secure()].
  factory XCBPrivateKey.createRandom(Random random) {
    generateNewPrivateKey(random).then((key) {
      return XCBPrivateKey(intToBytes(key));
    });
  }

  @override
  final bool isolateSafe = true;

  @override
  Future<XCBAddress> extractAddress() async {
    var _pubKey = await privateKeyBytesToPublic(privateKey);
    return _cachedAddress ??= XCBAddress(publicKeyToAddress(_pubKey));
  }

  @override
  Future<Uint8List> signToSignature(Uint8List payload, {int chainId}) async {
    print(bytesToHex(privateKey));
    final signature = await signWithPrivKey(sha3_256(payload), privateKey);
    // final chainIdV = 1
    // chainId != null ? (signature.v - 27 + (chainId * 2 + 35)) : signature.v;

    return signature;
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is XCBPrivateKey &&
          runtimeType == other.runtimeType &&
          const ListEquality().equals(privateKey, other.privateKey);

  @override
  int get hashCode => privateKey.hashCode;
}
